"""
Provides TIFFimage class.
"""
# Author: Pearu Peterson
# Created: June 2010


import os
import sys
import time
import numpy
import tif_lzw

from .utils import bytes2str
from .tiff_data import tag_name2value, tag_value2type, tag_value2name, \
    name2type, type2bytes, type2dtype

VERBOSE = True


class TIFFentry:
    """ Hold a IFD entry used by TIFFimage.
    """

    def __init__(self, tag):
        if isinstance(tag, str):
            tag = tag_name2value[tag]
        assert isinstance(tag, int), repr(tag)
        self.tag = tag
        self.type_name = tag_value2type[tag]
        self.type = name2type[self.type_name]
        self.type_nbytes = type2bytes[self.type]
        self.type_dtype = type2dtype[self.type]
        self.tag_name = tag_value2name.get(self.tag,
                                           'TAG%s' % (hex(self.tag),))

        self.record = numpy.zeros((12,), dtype=numpy.ubyte)
        self.record[:2].view(dtype=numpy.uint16)[0] = self.tag
        self.record[2:4].view(dtype=numpy.uint16)[0] = self.type
        self.values = []

    def __str__(self):
        return '%s(entry=(%s,%s,%s,%s))' % (self.__class__.__name__,
                                            self.tag_name,
                                            self.type_name,
                                            self.count,
                                            self.offset)

    __repr__ = __str__

    @property
    def count(self):
        return self.record[4:8].view(dtype=numpy.uint32)

    @property
    def offset(self):
        return self.record[8:12].view(dtype=numpy.uint32)

    @property
    def nbytes(self):
        if self.offset_is_value:
            return 0
        return self.count[0] * self.type_nbytes

    @property
    def offset_is_value(self):
        return not self.values and self.count[0] == 1 and \
               self.type_nbytes <= 4 and self.type_name != 'ASCII'

    def __getitem__(self, index):
        if self.offset_is_value:
            if index > 0:
                raise IndexError(repr(index))
            return self.offset[0]
        return self.values[index]

    def add_value(self, value):
        if isinstance(value, (list, tuple)):
            list(map(self.add_value, value))
        elif self.type_name == 'ASCII':
            value = str(value)
            if self.count[0] == 0:
                self.values.append(value)
            else:
                self.values[0] += value
            self.count[0] = len(self.values[0]) + 1
        elif self.type_nbytes <= 4:
            self.count[0] += 1
            if self.count[0] == 1:
                self.offset[0] = value
            elif self.count[0] == 2:
                self.values.append(self.offset[0])
                self.values.append(value)
                self.offset[0] = 0
            else:
                self.values.append(value)
        else:
            self.count[0] += 1
            self.values.append(value)

    def set_value(self, value):
        assert self.type_name != 'ASCII', repr(self)
        if self.count[0]:
            self.count[0] -= 1
            if self.values:
                del self.values[-1]
        self.add_value(value)

    def set_offset(self, offset):
        self.offset[0] = offset

    def toarray(self, target=None):
        if self.offset_is_value:
            return
        if target is None:
            target = numpy.zeros((self.nbytes,), dtype=numpy.ubyte)
        dtype = target.dtype
        offset = 0
        # print(self.values)
        # print(dtype)
        if self.type_name == 'ASCII':
            data = numpy.array([self.values[0] + '\0'])
            # print(type(data), data)
            data = numpy.array([self.values[0] + '\0'],
                               dtype='|S{}'.format(len(self.values[0]) +
                                                   1)).view(dtype=numpy.ubyte)
            # print(type(data), data)
            target[offset:offset + self.nbytes] = data
        else:
            for value in self.values:
                dtype = self.type_dtype
                if self.type_name == 'RATIONAL' and isinstance(value,
                                                               (int, float)):
                    dtype = numpy.float64
                target[offset:offset + self.type_nbytes].view(dtype=dtype)[
                    0] = value
                offset += self.type_nbytes
        return target


class TIFFimage:
    """
    Hold an image stack that can be written to TIFF file.
    """

    def __init__(self, data, description=''):
        """
        data : {list, numpy.ndarray}
          Specify image data as a list of images or as an array with rank<=3.
        """
        # dtype = None
        if isinstance(data, list):
            image = data[0]
            self.length, self.width = image.shape
            self.depth = len(data)
            dtype = image.dtype
        elif isinstance(data, numpy.ndarray):
            shape = data.shape
            dtype = data.dtype
            if len(shape) == 1:
                self.width, = shape
                self.length = 1
                self.depth = 1
                data = [[data]]
            elif len(shape) == 2:
                self.length, self.width = shape
                self.depth = 1
                data = [data]
            elif len(shape) == 3:
                self.depth, self.length, self.width = shape
            else:
                raise NotImplementedError(repr(shape))
        else:
            raise NotImplementedError(repr(type(data)))
        self.data = data
        self.dtype = dtype
        self.description = description

    # noinspection PyProtectedMember
    def write_file(self, filename, compression='none',
                   strip_size=2 ** 13, planar_config=1,
                   validate=False, verbose=None):
        """
        Write image data to TIFF file.

        Parameters
        ----------
        filename : str
        compression : {'none', 'lzw'}
        strip_size : int
          Specify the size of uncompressed strip.
        planar_config : int
        validate : bool
          When True then check compression by decompression.
        verbose : {bool, None}
          When True then write progress information to stdout. When None
          then verbose is assumed for data that has size over 1MB.

        Returns
        -------
        compression : float
          Compression factor.
        """
        if verbose is None:
            nbytes = self.depth * self.length * self.width * \
                     self.dtype.itemsize
            verbose = nbytes >= 1024 ** 2

        if os.path.splitext(filename)[1].lower() not in ['.tif', '.tiff']:
            filename += '.tif'

        if verbose:
            sys.stdout.write('Writing TIFF records to %s\n' % filename)
            sys.stdout.flush()

        compression_map = dict(packbits=32773, none=1, lzw=5, jpeg=6,
                               ccitt1d=2,
                               group3fax=3, group4fax=4
                               )
        compress_map = dict(none=lambda _data: _data,
                            lzw=tif_lzw.encode)
        decompress_map = dict(none=lambda _data, _bytes: _data,
                              lzw=tif_lzw.decode)
        compress = compress_map.get(compression or 'none', None)
        if compress is None:
            raise NotImplementedError(repr(compression))
        decompress = decompress_map.get(compression or 'none', None)
        # compute tif file size and create image file directories data
        image_directories = []
        total_size = 8
        data_size = 0
        image_data_size = 0
        for i, image in enumerate(self.data):
            if verbose:
                sys.stdout.write('\r  creating records: %5s%% done  ' % (
                    int(100.0 * i / len(self.data))))
                sys.stdout.flush()
            if image.dtype.kind == 'V' and len(
                    image.dtype.names) == 3:  # RGB image
                sample_format = dict(u=1, i=2, f=3, c=6).get(
                    image.dtype.fields[image.dtype.names[0]][0].kind)
                bits_per_sample = [image.dtype.fields[f][0].itemsize * 8 for f
                                   in image.dtype.names]
                samples_per_pixel = 3
                photometric_interpretation = 2
            else:  # gray scale image
                sample_format = dict(u=1, i=2, f=3, c=6).get(image.dtype.kind)
                bits_per_sample = image.dtype.itemsize * 8
                samples_per_pixel = 1
                photometric_interpretation = 1
            if sample_format is None:
                print('Warning(TIFFimage.write_file): unknown data kind %r, '
                      'mapping to void' % image.dtype.kind)
                sample_format = 4
            length, width = image.shape
            bytes_per_row = width * image.dtype.itemsize
            rows_per_strip = min(length,
                                 int(numpy.ceil(strip_size / bytes_per_row)))
            strips_per_image = int(
                numpy.floor((length + rows_per_strip - 1) / rows_per_strip))
            assert bytes_per_row * rows_per_strip * \
                strips_per_image >= image.nbytes
            d = dict(ImageWidth=width,
                     ImageLength=length,
                     Compression=compression_map.get(compression, 1),
                     PhotometricInterpretation=photometric_interpretation,
                     PlanarConfiguration=planar_config,
                     Orientation=1,
                     ResolutionUnit=1,
                     XResolution=1,
                     YResolution=1,
                     SamplesPerPixel=samples_per_pixel,
                     RowsPerStrip=rows_per_strip,
                     BitsPerSample=bits_per_sample,
                     SampleFormat=sample_format,
                     )
            if i == 0:
                d.update(dict(
                    ImageDescription=self.description,
                    Software='http://code.google.com/p/pylibtiff/'))

            entries = []
            for tagname, value in list(d.items()):
                entry = TIFFentry(tagname)
                entry.add_value(value)
                entries.append(entry)
                total_size += 12 + entry.nbytes
                data_size += entry.nbytes

            strip_byte_counts = TIFFentry('StripByteCounts')
            strip_offsets = TIFFentry('StripOffsets')
            entries.append(strip_byte_counts)
            entries.append(strip_offsets)
            # strip_offsets and strip_byte_counts will be filled in the next
            # loop
            if strips_per_image == 1:
                assert strip_byte_counts.type_nbytes <= 4
                assert strip_offsets.type_nbytes <= 4
                total_size += 2 * 12
            else:
                total_size += 2 * 12 + strips_per_image * (
                    strip_byte_counts.type_nbytes + strip_offsets.type_nbytes)
                data_size += strips_per_image * (
                    strip_byte_counts.type_nbytes + strip_offsets.type_nbytes)

            # image data:
            total_size += image.nbytes
            data_size += image.nbytes
            image_data_size += image.nbytes

            # records for nof IFD entries and offset to the next IFD:
            total_size += 2 + 4

            # entries must be sorted by tag number
            entries.sort(key=lambda x: x.tag)

            strip_info = strip_offsets, strip_byte_counts, strips_per_image,\
                rows_per_strip, bytes_per_row
            image_directories.append((entries, strip_info, image))

        tif = numpy.memmap(filename, dtype=numpy.ubyte, mode='w+',
                           shape=(total_size,))

        # noinspection PyProtectedMember
        def tif_write(_tif, _offset, _data):
            end = _offset + _data.nbytes
            if end > _tif.size:
                size_incr = int((end - _tif.size) / 1024 ** 2 + 1) * 1024 ** 2
                new_size = _tif.size + size_incr
                assert end <= new_size, repr(
                    (end, _tif.size, size_incr, new_size))
                # sys.stdout.write('resizing: %s -> %s\n' % (tif.size,
                # new_size))
                # tif.resize(end, refcheck=False)
                _base = _tif._mmap
                if _base is None:
                    _base = _tif.base
                _base.resize(new_size)
                new_tif = numpy.ndarray.__new__(numpy.memmap, (_base.size(),),
                                                dtype=_tif.dtype, buffer=_base)
                new_tif._parent = _tif
                new_tif.__array_finalize__(_tif)
                _tif = new_tif
            _tif[_offset:end] = _data
            return _tif

        # write TIFF header
        tif[:2].view(dtype=numpy.uint16)[0] = 0x4949  # low-endian
        tif[2:4].view(dtype=numpy.uint16)[0] = 42  # magic number
        tif[4:8].view(dtype=numpy.uint32)[0] = 8  # offset to the first IFD

        offset = 8
        data_offset = total_size - data_size
        image_data_offset = total_size - image_data_size
        first_data_offset = data_offset
        first_image_data_offset = image_data_offset
        start_time = time.time()
        compressed_data_size = 0
        for i, (entries, strip_info, image) in enumerate(image_directories):
            strip_offsets, strip_byte_counts, strips_per_image, rows_per_strip, bytes_per_row = strip_info

            # write the nof IFD entries
            tif[offset:offset + 2].view(dtype=numpy.uint16)[0] = len(entries)
            offset += 2
            assert offset <= first_data_offset, repr(
                (offset, first_data_offset))

            # write image data
            data = image.view(dtype=numpy.ubyte).reshape((image.nbytes,))

            for j in range(strips_per_image):
                c = rows_per_strip * bytes_per_row
                k = j * c
                c -= max((j + 1) * c - image.nbytes, 0)
                assert c > 0, repr(c)
                orig_strip = data[k:k + c]  # type: numpy.ndarray

                strip = compress(orig_strip)
                if validate:
                    test_strip = decompress(strip, orig_strip.nbytes)
                    if (orig_strip != test_strip).any():
                        raise RuntimeError(
                            'Compressed data is corrupted: cannot recover '
                            'original data')
                compressed_data_size += strip.nbytes
                # print strip.size, strip.nbytes, strip.shape,
                # tif[image_data_offset:image_data_offset+strip.nbytes].shape
                strip_offsets.add_value(image_data_offset)
                strip_byte_counts.add_value(strip.nbytes)

                tif = tif_write(tif, image_data_offset, strip)
                image_data_offset += strip.nbytes
                # if j == 0:
                #     first = strip_offsets[0]
                # last = strip_offsets[-1] + strip_byte_counts[-1]

            # write IFD entries
            for entry in entries:
                data_size = entry.nbytes
                if data_size:
                    entry.set_offset(data_offset)
                    assert data_offset + data_size <= total_size, repr(
                        (data_offset + data_size, total_size))
                    r = entry.toarray(tif[data_offset:data_offset + data_size])
                    assert r.nbytes == data_size
                    data_offset += data_size
                    assert data_offset <= first_image_data_offset, repr(
                        (data_offset, first_image_data_offset, i))
                tif[offset:offset + 12] = entry.record
                offset += 12
                assert offset <= first_data_offset, repr(
                    (offset, first_data_offset, i))

            # write offset to the next IFD
            tif[offset:offset + 4].view(dtype=numpy.uint32)[0] = offset + 4
            offset += 4
            assert offset <= first_data_offset, repr(
                (offset, first_data_offset))

            if verbose:
                sys.stdout.write(
                    '\r  filling records: %5s%% done (%s/s)%s' %
                    (int(100.0 * (i + 1) / len(image_directories)),
                     bytes2str(int((image_data_offset -
                                    first_image_data_offset) /
                                   (time.time() - start_time))), ' ' * 2))
                if (i + 1) == len(image_directories):
                    sys.stdout.write('\n')
                sys.stdout.flush()

        # last offset must be 0
        tif[offset - 4:offset].view(dtype=numpy.uint32)[0] = 0

        compression = 1 / (compressed_data_size / image_data_size)

        if compressed_data_size != image_data_size:
            sdiff = image_data_size - compressed_data_size
            total_size -= sdiff
            base = tif._mmap
            if base is None:
                base = tif.base
            base.resize(total_size)
            if verbose:
                sys.stdout.write(
                    '  resized records: %s -> %s (compression: %.2fx)\n'
                    % (bytes2str(total_size + sdiff), bytes2str(total_size),
                       compression))
                sys.stdout.flush()
        del tif  # flushing
        return compression



"""
tiff - implements a numpy.memmap based TIFF file reader and writer
allowing manipulating TIFF files that have sizes larger than
available memory in computer.

Usage:
>>> tiff = TIFFfile('<filename.(tif|lsm)>')
>>> samples, sample_names = tiff.get_samples()
>>> arr = tiff.get_tiff_array(sample_index=0, subfile_type=0)

>>> tiff = TIFFimage(data, description=<str>)
>>> tiff.write_file (<filename.tif>, compression='none'|'lzw')
>>> del tiff # flush data to disk
"""
# Author: Pearu Peterson
# Created: April 2010

__all__ = ['TIFFfile', 'TIFFimage', 'TiffArray']

import os
import sys
import time
import numpy

from .tiff_file import TIFFfile
from .tiff_image import TIFFimage
from .tiff_array import TiffArray

def main ():
    filename = sys.argv[1]
    if not os.path.isfile(filename):
        raise ValueError('File %r does not exists' % (filename))

    t = TIFFfile(filename)

    t.show_memory_usage()

    e = t.IFD[0].entries[-1]
    assert e.is_lsm
    from . import lsm
    print(lsm.lsmblock(e))
    print(lsm.lsminfo(e, 0))
    #print lsm.filestructure(e)
    #print lsm.timestamps(e)
    #print lsm.channelwavelength(e)

if __name__ == '__main__':
    main()





from .tiff_base import TiffBase

class TiffChannelsAndFiles (TiffBase):
    """ Represent a collection of TIFF files as a single TIFF source object.

    See also
    --------
    TiffFile, TiffFiles
    """

    def __init__(self, channels_files_map):
        """
        Parameters
        ----------
        channels_files_map : dict
          A dictionary of channel names and TIFF files (``TiffFiles`` instances)
        """
        self.channels_files_map = channels_files_map
    
    def get_tiff_array (self, channel, sample_index=0, subfile_type=0, assume_one_image_per_file=False):
        """ Return an array of images for given channel.

        Parameters
        ----------
        channel : str
          The name of a channel.
        sample_index : int
          Specify sample within a pixel.
        subfile_type : int
          Specify TIFF NewSubfileType used for collecting sample images.
        assume_one_image_per_file : bool
          When True then it is assumed that each TIFF file contains
          exactly one image and all images have the same parameters.
          This knowledge speeds up tiff_array construction as only the
          first TIFF file is opened for reading image parameters. The
          other TIFF files are opened only when particular images are
          accessed.

        Returns
        -------
        tiff_array : TiffArray
          Array of sample images. The array has rank equal to 3.
        """
        return self.channels_files_map[channel].get_tiff_array(sample_index=sample_index,
                                                               subfile_type=subfile_type,
                                                               assume_one_image_per_file=assume_one_image_per_file)

    def get_info(self):
        l = []
        for channel, tiff in list(self.channels_files_map.items ()):
            l.append ('Channel %s:' % (channel))
            l.append ('-'*len(l[-1]))
            l.append (tiff.get_info())
        return '\n'.join(l)

    def close(self):
        for channel, tiff in list(self.channels_files_map.items()):
            tiff.close()



TIFF_VERSION_CLASSIC = 42
TIFF_VERSION_BIG = 43
TIFF_BIGENDIAN = 19789
TIFF_LITTLEENDIAN = 18761
MDI_LITTLEENDIAN = 20549
MDI_BIGENDIAN = 17744
TIFFTAG_SUBFILETYPE = 254
FILETYPE_REDUCEDIMAGE = 1
FILETYPE_PAGE = 2
FILETYPE_MASK = 4
TIFFTAG_OSUBFILETYPE = 255
OFILETYPE_IMAGE = 1
OFILETYPE_REDUCEDIMAGE = 2
OFILETYPE_PAGE = 3
TIFFTAG_IMAGEWIDTH = 256
TIFFTAG_IMAGELENGTH = 257
TIFFTAG_BITSPERSAMPLE = 258
TIFFTAG_COMPRESSION = 259
COMPRESSION_NONE = 1
COMPRESSION_CCITTRLE = 2
COMPRESSION_CCITTFAX3 = 3
COMPRESSION_CCITT_T4 = 3
COMPRESSION_CCITTFAX4 = 4
COMPRESSION_CCITT_T6 = 4
COMPRESSION_LZW = 5
COMPRESSION_OJPEG = 6
COMPRESSION_JPEG = 7
COMPRESSION_T85 = 9
COMPRESSION_T43 = 10
COMPRESSION_NEXT = 32766
COMPRESSION_CCITTRLEW = 32771
COMPRESSION_PACKBITS = 32773
COMPRESSION_THUNDERSCAN = 32809
COMPRESSION_IT8CTPAD = 32895
COMPRESSION_IT8LW = 32896
COMPRESSION_IT8MP = 32897
COMPRESSION_IT8BL = 32898
COMPRESSION_PIXARFILM = 32908
COMPRESSION_PIXARLOG = 32909
COMPRESSION_DEFLATE = 32946
COMPRESSION_ADOBE_DEFLATE = 8
COMPRESSION_DCS = 32947
COMPRESSION_JBIG = 34661
COMPRESSION_SGILOG = 34676
COMPRESSION_SGILOG24 = 34677
COMPRESSION_JP2000 = 34712
COMPRESSION_LERC = 34887
COMPRESSION_LZMA = 34925
COMPRESSION_ZSTD = 50000
COMPRESSION_WEBP = 50001
TIFFTAG_PHOTOMETRIC = 262
PHOTOMETRIC_MINISWHITE = 0
PHOTOMETRIC_MINISBLACK = 1
PHOTOMETRIC_RGB = 2
PHOTOMETRIC_PALETTE = 3
PHOTOMETRIC_MASK = 4
PHOTOMETRIC_SEPARATED = 5
PHOTOMETRIC_YCBCR = 6
PHOTOMETRIC_CIELAB = 8
PHOTOMETRIC_ICCLAB = 9
PHOTOMETRIC_ITULAB = 10
PHOTOMETRIC_CFA = 32803
PHOTOMETRIC_LOGL = 32844
PHOTOMETRIC_LOGLUV = 32845
TIFFTAG_THRESHHOLDING = 263
THRESHHOLD_BILEVEL = 1
THRESHHOLD_HALFTONE = 2
THRESHHOLD_ERRORDIFFUSE = 3
TIFFTAG_CELLWIDTH = 264
TIFFTAG_CELLLENGTH = 265
TIFFTAG_FILLORDER = 266
FILLORDER_MSB2LSB = 1
FILLORDER_LSB2MSB = 2
TIFFTAG_DOCUMENTNAME = 269
TIFFTAG_IMAGEDESCRIPTION = 270
TIFFTAG_MAKE = 271
TIFFTAG_MODEL = 272
TIFFTAG_STRIPOFFSETS = 273
TIFFTAG_ORIENTATION = 274
ORIENTATION_TOPLEFT = 1
ORIENTATION_TOPRIGHT = 2
ORIENTATION_BOTRIGHT = 3
ORIENTATION_BOTLEFT = 4
ORIENTATION_LEFTTOP = 5
ORIENTATION_RIGHTTOP = 6
ORIENTATION_RIGHTBOT = 7
ORIENTATION_LEFTBOT = 8
TIFFTAG_SAMPLESPERPIXEL = 277
TIFFTAG_ROWSPERSTRIP = 278
TIFFTAG_STRIPBYTECOUNTS = 279
TIFFTAG_MINSAMPLEVALUE = 280
TIFFTAG_MAXSAMPLEVALUE = 281
TIFFTAG_XRESOLUTION = 282
TIFFTAG_YRESOLUTION = 283
TIFFTAG_PLANARCONFIG = 284
PLANARCONFIG_CONTIG = 1
PLANARCONFIG_SEPARATE = 2
TIFFTAG_PAGENAME = 285
TIFFTAG_XPOSITION = 286
TIFFTAG_YPOSITION = 287
TIFFTAG_FREEOFFSETS = 288
TIFFTAG_FREEBYTECOUNTS = 289
TIFFTAG_GRAYRESPONSEUNIT = 290
GRAYRESPONSEUNIT_10S = 1
GRAYRESPONSEUNIT_100S = 2
GRAYRESPONSEUNIT_1000S = 3
GRAYRESPONSEUNIT_10000S = 4
GRAYRESPONSEUNIT_100000S = 5
TIFFTAG_GRAYRESPONSECURVE = 291
TIFFTAG_GROUP3OPTIONS = 292
TIFFTAG_T4OPTIONS = 292
GROUP3OPT_2DENCODING = 1
GROUP3OPT_UNCOMPRESSED = 2
GROUP3OPT_FILLBITS = 4
TIFFTAG_GROUP4OPTIONS = 293
TIFFTAG_T6OPTIONS = 293
GROUP4OPT_UNCOMPRESSED = 2
TIFFTAG_RESOLUTIONUNIT = 296
RESUNIT_NONE = 1
RESUNIT_INCH = 2
RESUNIT_CENTIMETER = 3
TIFFTAG_PAGENUMBER = 297
TIFFTAG_COLORRESPONSEUNIT = 300
COLORRESPONSEUNIT_10S = 1
COLORRESPONSEUNIT_100S = 2
COLORRESPONSEUNIT_1000S = 3
COLORRESPONSEUNIT_10000S = 4
COLORRESPONSEUNIT_100000S = 5
TIFFTAG_TRANSFERFUNCTION = 301
TIFFTAG_SOFTWARE = 305
TIFFTAG_DATETIME = 306
TIFFTAG_ARTIST = 315
TIFFTAG_HOSTCOMPUTER = 316
TIFFTAG_PREDICTOR = 317
PREDICTOR_NONE = 1
PREDICTOR_HORIZONTAL = 2
PREDICTOR_FLOATINGPOINT = 3
TIFFTAG_WHITEPOINT = 318
TIFFTAG_PRIMARYCHROMATICITIES = 319
TIFFTAG_COLORMAP = 320
TIFFTAG_HALFTONEHINTS = 321
TIFFTAG_TILEWIDTH = 322
TIFFTAG_TILELENGTH = 323
TIFFTAG_TILEOFFSETS = 324
TIFFTAG_TILEBYTECOUNTS = 325
TIFFTAG_BADFAXLINES = 326
TIFFTAG_CLEANFAXDATA = 327
CLEANFAXDATA_CLEAN = 0
CLEANFAXDATA_REGENERATED = 1
CLEANFAXDATA_UNCLEAN = 2
TIFFTAG_CONSECUTIVEBADFAXLINES = 328
TIFFTAG_SUBIFD = 330
TIFFTAG_INKSET = 332
INKSET_CMYK = 1
INKSET_MULTIINK = 2
TIFFTAG_INKNAMES = 333
TIFFTAG_NUMBEROFINKS = 334
TIFFTAG_DOTRANGE = 336
TIFFTAG_TARGETPRINTER = 337
TIFFTAG_EXTRASAMPLES = 338
EXTRASAMPLE_UNSPECIFIED = 0
EXTRASAMPLE_ASSOCALPHA = 1
EXTRASAMPLE_UNASSALPHA = 2
TIFFTAG_SAMPLEFORMAT = 339
SAMPLEFORMAT_UINT = 1
SAMPLEFORMAT_INT = 2
SAMPLEFORMAT_IEEEFP = 3
SAMPLEFORMAT_VOID = 4
SAMPLEFORMAT_COMPLEXINT = 5
SAMPLEFORMAT_COMPLEXIEEEFP = 6
TIFFTAG_SMINSAMPLEVALUE = 340
TIFFTAG_SMAXSAMPLEVALUE = 341
TIFFTAG_CLIPPATH = 343
TIFFTAG_XCLIPPATHUNITS = 344
TIFFTAG_YCLIPPATHUNITS = 345
TIFFTAG_INDEXED = 346
TIFFTAG_JPEGTABLES = 347
TIFFTAG_OPIPROXY = 351
TIFFTAG_GLOBALPARAMETERSIFD = 400
TIFFTAG_PROFILETYPE = 401
PROFILETYPE_UNSPECIFIED = 0
PROFILETYPE_G3_FAX = 1
TIFFTAG_FAXPROFILE = 402
FAXPROFILE_S = 1
FAXPROFILE_F = 2
FAXPROFILE_J = 3
FAXPROFILE_C = 4
FAXPROFILE_L = 5
FAXPROFILE_M = 6
TIFFTAG_CODINGMETHODS = 403
CODINGMETHODS_T4_1D = 2
CODINGMETHODS_T4_2D = 4
CODINGMETHODS_T6 = 8
CODINGMETHODS_T85 = 16
CODINGMETHODS_T42 = 32
CODINGMETHODS_T43 = 64
TIFFTAG_VERSIONYEAR = 404
TIFFTAG_MODENUMBER = 405
TIFFTAG_DECODE = 433
TIFFTAG_IMAGEBASECOLOR = 434
TIFFTAG_T82OPTIONS = 435
TIFFTAG_JPEGPROC = 512
JPEGPROC_BASELINE = 1
JPEGPROC_LOSSLESS = 14
TIFFTAG_JPEGIFOFFSET = 513
TIFFTAG_JPEGIFBYTECOUNT = 514
TIFFTAG_JPEGRESTARTINTERVAL = 515
TIFFTAG_JPEGLOSSLESSPREDICTORS = 517
TIFFTAG_JPEGPOINTTRANSFORM = 518
TIFFTAG_JPEGQTABLES = 519
TIFFTAG_JPEGDCTABLES = 520
TIFFTAG_JPEGACTABLES = 521
TIFFTAG_YCBCRCOEFFICIENTS = 529
TIFFTAG_YCBCRSUBSAMPLING = 530
TIFFTAG_YCBCRPOSITIONING = 531
YCBCRPOSITION_CENTERED = 1
YCBCRPOSITION_COSITED = 2
TIFFTAG_REFERENCEBLACKWHITE = 532
TIFFTAG_STRIPROWCOUNTS = 559
TIFFTAG_XMLPACKET = 700
TIFFTAG_OPIIMAGEID = 32781
TIFFTAG_REFPTS = 32953
TIFFTAG_REGIONTACKPOINT = 32954
TIFFTAG_REGIONWARPCORNERS = 32955
TIFFTAG_REGIONAFFINE = 32956
TIFFTAG_MATTEING = 32995
TIFFTAG_DATATYPE = 32996
TIFFTAG_IMAGEDEPTH = 32997
TIFFTAG_TILEDEPTH = 32998
TIFFTAG_PIXAR_IMAGEFULLWIDTH = 33300
TIFFTAG_PIXAR_IMAGEFULLLENGTH = 33301
TIFFTAG_PIXAR_TEXTUREFORMAT = 33302
TIFFTAG_PIXAR_WRAPMODES = 33303
TIFFTAG_PIXAR_FOVCOT = 33304
TIFFTAG_PIXAR_MATRIX_WORLDTOSCREEN = 33305
TIFFTAG_PIXAR_MATRIX_WORLDTOCAMERA = 33306
TIFFTAG_WRITERSERIALNUMBER = 33405
TIFFTAG_CFAREPEATPATTERNDIM = 33421
TIFFTAG_CFAPATTERN = 33422
TIFFTAG_COPYRIGHT = 33432
TIFFTAG_RICHTIFFIPTC = 33723
TIFFTAG_IT8SITE = 34016
TIFFTAG_IT8COLORSEQUENCE = 34017
TIFFTAG_IT8HEADER = 34018
TIFFTAG_IT8RASTERPADDING = 34019
TIFFTAG_IT8BITSPERRUNLENGTH = 34020
TIFFTAG_IT8BITSPEREXTENDEDRUNLENGTH = 34021
TIFFTAG_IT8COLORTABLE = 34022
TIFFTAG_IT8IMAGECOLORINDICATOR = 34023
TIFFTAG_IT8BKGCOLORINDICATOR = 34024
TIFFTAG_IT8IMAGECOLORVALUE = 34025
TIFFTAG_IT8BKGCOLORVALUE = 34026
TIFFTAG_IT8PIXELINTENSITYRANGE = 34027
TIFFTAG_IT8TRANSPARENCYINDICATOR = 34028
TIFFTAG_IT8COLORCHARACTERIZATION = 34029
TIFFTAG_IT8HCUSAGE = 34030
TIFFTAG_IT8TRAPINDICATOR = 34031
TIFFTAG_IT8CMYKEQUIVALENT = 34032
TIFFTAG_FRAMECOUNT = 34232
TIFFTAG_PHOTOSHOP = 34377
TIFFTAG_EXIFIFD = 34665
TIFFTAG_ICCPROFILE = 34675
TIFFTAG_IMAGELAYER = 34732
TIFFTAG_JBIGOPTIONS = 34750
TIFFTAG_GPSIFD = 34853
TIFFTAG_FAXRECVPARAMS = 34908
TIFFTAG_FAXSUBADDRESS = 34909
TIFFTAG_FAXRECVTIME = 34910
TIFFTAG_FAXDCS = 34911
TIFFTAG_STONITS = 37439
TIFFTAG_FEDEX_EDR = 34929
TIFFTAG_INTEROPERABILITYIFD = 40965
TIFFTAG_LERC_PARAMETERS = 50674
TIFFTAG_DNGVERSION = 50706
TIFFTAG_DNGBACKWARDVERSION = 50707
TIFFTAG_UNIQUECAMERAMODEL = 50708
TIFFTAG_LOCALIZEDCAMERAMODEL = 50709
TIFFTAG_CFAPLANECOLOR = 50710
TIFFTAG_CFALAYOUT = 50711
TIFFTAG_LINEARIZATIONTABLE = 50712
TIFFTAG_BLACKLEVELREPEATDIM = 50713
TIFFTAG_BLACKLEVEL = 50714
TIFFTAG_BLACKLEVELDELTAH = 50715
TIFFTAG_BLACKLEVELDELTAV = 50716
TIFFTAG_WHITELEVEL = 50717
TIFFTAG_DEFAULTSCALE = 50718
TIFFTAG_DEFAULTCROPORIGIN = 50719
TIFFTAG_DEFAULTCROPSIZE = 50720
TIFFTAG_COLORMATRIX1 = 50721
TIFFTAG_COLORMATRIX2 = 50722
TIFFTAG_CAMERACALIBRATION1 = 50723
TIFFTAG_CAMERACALIBRATION2 = 50724
TIFFTAG_REDUCTIONMATRIX1 = 50725
TIFFTAG_REDUCTIONMATRIX2 = 50726
TIFFTAG_ANALOGBALANCE = 50727
TIFFTAG_ASSHOTNEUTRAL = 50728
TIFFTAG_ASSHOTWHITEXY = 50729
TIFFTAG_BASELINEEXPOSURE = 50730
TIFFTAG_BASELINENOISE = 50731
TIFFTAG_BASELINESHARPNESS = 50732
TIFFTAG_BAYERGREENSPLIT = 50733
TIFFTAG_LINEARRESPONSELIMIT = 50734
TIFFTAG_CAMERASERIALNUMBER = 50735
TIFFTAG_LENSINFO = 50736
TIFFTAG_CHROMABLURRADIUS = 50737
TIFFTAG_ANTIALIASSTRENGTH = 50738
TIFFTAG_SHADOWSCALE = 50739
TIFFTAG_DNGPRIVATEDATA = 50740
TIFFTAG_MAKERNOTESAFETY = 50741
TIFFTAG_CALIBRATIONILLUMINANT1 = 50778
TIFFTAG_CALIBRATIONILLUMINANT2 = 50779
TIFFTAG_BESTQUALITYSCALE = 50780
TIFFTAG_RAWDATAUNIQUEID = 50781
TIFFTAG_ORIGINALRAWFILENAME = 50827
TIFFTAG_ORIGINALRAWFILEDATA = 50828
TIFFTAG_ACTIVEAREA = 50829
TIFFTAG_MASKEDAREAS = 50830
TIFFTAG_ASSHOTICCPROFILE = 50831
TIFFTAG_ASSHOTPREPROFILEMATRIX = 50832
TIFFTAG_CURRENTICCPROFILE = 50833
TIFFTAG_CURRENTPREPROFILEMATRIX = 50834
TIFFTAG_DCSHUESHIFTVALUES = 65535
TIFFTAG_FAXMODE = 65536
FAXMODE_CLASSIC = 0
FAXMODE_NORTC = 1
FAXMODE_NOEOL = 2
FAXMODE_BYTEALIGN = 4
FAXMODE_WORDALIGN = 8
FAXMODE_CLASSF = 1
TIFFTAG_JPEGQUALITY = 65537
TIFFTAG_JPEGCOLORMODE = 65538
JPEGCOLORMODE_RAW = 0
JPEGCOLORMODE_RGB = 1
TIFFTAG_JPEGTABLESMODE = 65539
JPEGTABLESMODE_QUANT = 1
JPEGTABLESMODE_HUFF = 2
TIFFTAG_FAXFILLFUNC = 65540
TIFFTAG_PIXARLOGDATAFMT = 65549
PIXARLOGDATAFMT_8BIT = 0
PIXARLOGDATAFMT_8BITABGR = 1
PIXARLOGDATAFMT_11BITLOG = 2
PIXARLOGDATAFMT_12BITPICIO = 3
PIXARLOGDATAFMT_16BIT = 4
PIXARLOGDATAFMT_FLOAT = 5
TIFFTAG_DCSIMAGERTYPE = 65550
DCSIMAGERMODEL_M3 = 0
DCSIMAGERMODEL_M5 = 1
DCSIMAGERMODEL_M6 = 2
DCSIMAGERFILTER_IR = 0
DCSIMAGERFILTER_MONO = 1
DCSIMAGERFILTER_CFA = 2
DCSIMAGERFILTER_OTHER = 3
TIFFTAG_DCSINTERPMODE = 65551
DCSINTERPMODE_NORMAL = 0
DCSINTERPMODE_PREVIEW = 1
TIFFTAG_DCSBALANCEARRAY = 65552
TIFFTAG_DCSCORRECTMATRIX = 65553
TIFFTAG_DCSGAMMA = 65554
TIFFTAG_DCSTOESHOULDERPTS = 65555
TIFFTAG_DCSCALIBRATIONFD = 65556
TIFFTAG_ZIPQUALITY = 65557
TIFFTAG_PIXARLOGQUALITY = 65558
TIFFTAG_DCSCLIPRECTANGLE = 65559
TIFFTAG_SGILOGDATAFMT = 65560
SGILOGDATAFMT_FLOAT = 0
SGILOGDATAFMT_16BIT = 1
SGILOGDATAFMT_RAW = 2
SGILOGDATAFMT_8BIT = 3
TIFFTAG_SGILOGENCODE = 65561
SGILOGENCODE_NODITHER = 0
SGILOGENCODE_RANDITHER = 1
TIFFTAG_LZMAPRESET = 65562
TIFFTAG_PERSAMPLE = 65563
PERSAMPLE_MERGED = 0
PERSAMPLE_MULTI = 1
TIFFTAG_ZSTD_LEVEL = 65564
TIFFTAG_LERC_VERSION = 65565
LERC_VERSION_2_4 = 4
TIFFTAG_LERC_ADD_COMPRESSION = 65566
LERC_ADD_COMPRESSION_NONE = 0
LERC_ADD_COMPRESSION_DEFLATE = 1
LERC_ADD_COMPRESSION_ZSTD = 2
TIFFTAG_LERC_MAXZERROR = 65567
TIFFTAG_WEBP_LEVEL = 65568
TIFFTAG_WEBP_LOSSLESS = 65569
EXIFTAG_EXPOSURETIME = 33434
EXIFTAG_FNUMBER = 33437
EXIFTAG_EXPOSUREPROGRAM = 34850
EXIFTAG_SPECTRALSENSITIVITY = 34852
EXIFTAG_ISOSPEEDRATINGS = 34855
EXIFTAG_OECF = 34856
EXIFTAG_EXIFVERSION = 36864
EXIFTAG_DATETIMEORIGINAL = 36867
EXIFTAG_DATETIMEDIGITIZED = 36868
EXIFTAG_COMPONENTSCONFIGURATION = 37121
EXIFTAG_COMPRESSEDBITSPERPIXEL = 37122
EXIFTAG_SHUTTERSPEEDVALUE = 37377
EXIFTAG_APERTUREVALUE = 37378
EXIFTAG_BRIGHTNESSVALUE = 37379
EXIFTAG_EXPOSUREBIASVALUE = 37380
EXIFTAG_MAXAPERTUREVALUE = 37381
EXIFTAG_SUBJECTDISTANCE = 37382
EXIFTAG_METERINGMODE = 37383
EXIFTAG_LIGHTSOURCE = 37384
EXIFTAG_FLASH = 37385
EXIFTAG_FOCALLENGTH = 37386
EXIFTAG_SUBJECTAREA = 37396
EXIFTAG_MAKERNOTE = 37500
EXIFTAG_USERCOMMENT = 37510
EXIFTAG_SUBSECTIME = 37520
EXIFTAG_SUBSECTIMEORIGINAL = 37521
EXIFTAG_SUBSECTIMEDIGITIZED = 37522
EXIFTAG_FLASHPIXVERSION = 40960
EXIFTAG_COLORSPACE = 40961
EXIFTAG_PIXELXDIMENSION = 40962
EXIFTAG_PIXELYDIMENSION = 40963
EXIFTAG_RELATEDSOUNDFILE = 40964
EXIFTAG_FLASHENERGY = 41483
EXIFTAG_SPATIALFREQUENCYRESPONSE = 41484
EXIFTAG_FOCALPLANEXRESOLUTION = 41486
EXIFTAG_FOCALPLANEYRESOLUTION = 41487
EXIFTAG_FOCALPLANERESOLUTIONUNIT = 41488
EXIFTAG_SUBJECTLOCATION = 41492
EXIFTAG_EXPOSUREINDEX = 41493
EXIFTAG_SENSINGMETHOD = 41495
EXIFTAG_FILESOURCE = 41728
EXIFTAG_SCENETYPE = 41729
EXIFTAG_CFAPATTERN = 41730
EXIFTAG_CUSTOMRENDERED = 41985
EXIFTAG_EXPOSUREMODE = 41986
EXIFTAG_WHITEBALANCE = 41987
EXIFTAG_DIGITALZOOMRATIO = 41988
EXIFTAG_FOCALLENGTHIN35MMFILM = 41989
EXIFTAG_SCENECAPTURETYPE = 41990
EXIFTAG_GAINCONTROL = 41991
EXIFTAG_CONTRAST = 41992
EXIFTAG_SATURATION = 41993
EXIFTAG_SHARPNESS = 41994
EXIFTAG_DEVICESETTINGDESCRIPTION = 41995
EXIFTAG_SUBJECTDISTANCERANGE = 41996
EXIFTAG_GAINCONTROL = 41991
EXIFTAG_GAINCONTROL = 41991
EXIFTAG_IMAGEUNIQUEID = 42016




import sys
from os.path import join, basename, dirname, splitext
from glob import glob

from numpy.distutils import log
from distutils.dep_util import newer


def configuration(parent_package='',top_path=None):
    from numpy.distutils.misc_util import Configuration
    package_name = 'libtiff'
    config = Configuration(package_name,parent_package,top_path)

    bitarray_path = 'bitarray-a1646c0/bitarray'

    # Add subpackages here:
    config.add_subpackage('bitarray', bitarray_path)
    # eof add.

    # Add extensions here:
    config.add_extension('bitarray._bitarray', join(bitarray_path,'_bitarray.c'))
    config.add_extension('bittools', join('src','bittools.c'))
    config.add_extension('tif_lzw', join('src','tif_lzw.c'))
    # eof add.

    config.make_svn_version_py()

    wininst = 'bdist_wininst' in sys.argv

    # Scripts support: files in scripts directories are considered as
    # python scripts that will be installed as
    # <package_name>.<script_name> to scripts installation directory.
    scripts = glob(join(config.local_path, 'scripts', '*.py'))
    scripts += glob(join(config.local_path, '*', 'scripts', '*.py'))
    for script in scripts:
        if basename (script).startswith (package_name):
            config.add_scripts(script)
            continue

        def generate_a_script(build_dir, script=script, config=config):
            dist = config.get_distribution()
            install_lib = dist.get_command_obj('install_lib')
            if not install_lib.finalized:
                install_lib.finalize_options()

            script_replace_text = ''
            install_lib = install_lib.install_dir
            if install_lib is not None:
                script_replace_text = '''
import sys
if %(d)r not in sys.path:
    sys.path.insert(0, %(d)r)
''' % dict(d=install_lib)

            start_mark = '### START UPDATE SYS.PATH ###'
            end_mark = '### END UPDATE SYS.PATH ###'
            name = basename(script)
            if name.startswith (package_name):
                target_name = name
            elif wininst:
                target_name = package_name + '_' + name
            else:
                target_name = package_name + '.' + splitext(name)[0]
            target = join(build_dir, target_name)
            if newer(script, target) or 1:
                log.info('Creating %r', target)
                f = open (script, 'r')
                text = f.read()
                f.close()

                i = text.find(start_mark)
                if i != -1:
                    j = text.find (end_mark)
                    if j == -1:
                        log.warn ("%r missing %r line", script, start_mark)
                    new_text = text[:i+len (start_mark)] + script_replace_text + text[j:]
                else:
                    new_text = text

                f = open(target, 'w')
                f.write(new_text)
                f.close()
            print("*****{0}".format(target_name))
        config.add_scripts(generate_a_script)
        

    return config

if __name__ == '__main__':
    from numpy.distutils.core import setup
    setup (configuration=configuration)



""" Encoder and decoder of Lempel-Ziv-Welch algorithm for TIFF.

This module is obsolete, use tif_lzw extension module instead.
"""
# Author: Pearu Peterson
# Created: May 2010

import numpy

default_backend='bittools'
#default_backend='bittools'

if default_backend=='bitarray':
    from bitarray import bitarray
if default_backend=='bittools':
    from bittools import setword, getword

CODECLEAR = 256
CODEEOI = 257
CODESTART = 258

def encode_bitarray(seq, max_bits=12):
    """ Compress sequence using Lempel-Ziv-Welch algorithm for TIFF.

    Parameters
    ----------
    seq : {str, numpy.ndarray}
    max_bits : int
      Specify maximum bits for encoding table.

    Returns
    -------
    bseq : bitarray

    See also
    --------
    decode_bitarray
    """
    if isinstance (seq, numpy.ndarray):
        seq = seq.tostring()
    r = bitarray(0, endian='little')
    write = r.fromword

    init_table = [(chr(code),code) for code in range (256)]

    table = {}
    table_get = table.get
    table_clear = table.clear
    table_update = table.update

    sup_code2 = (1<<max_bits) - 2
    next_code = CODESTART
    bits = 9
    max_code = (1<<bits)
    s = ''

    table_update(init_table)

    index = 0
    write(CODECLEAR, bits)
    for c in seq:
        s1 = s + c
        if s1 in table:
            s = s1
        else:
            write(table_get(s), bits)
            table[s1] = next_code
            next_code += 1
            s = c
            if next_code==sup_code2:
                write(table_get(s), bits)
                write(CODECLEAR, bits)
                s = ''
                table_clear()
                table_update(init_table)
                next_code = CODESTART
                bits = 9
                max_code = (1<<bits)
            elif next_code==max_code:
                bits += 1
                max_code = (1<<bits)
    if s:
        write(table_get(s), bits)
    write(CODEEOI, bits)
    return r

def encode_bittools(seq, max_bits=12):
    """ Compress sequence using Lempel-Ziv-Welch algorithm for TIFF.

    Parameters
    ----------
    seq : {str, numpy.ndarray}
    max_bits : int
      Specify maximum bits for encoding table.

    Returns
    -------
    bseq : numpy.ndarray

    See also
    --------
    decode_bittools
    """
    if isinstance (seq, numpy.ndarray):
        nbytes = seq.nbytes*2
        seq = seq.tostring()
    else:
        nbytes = len(seq)*2
    r = numpy.zeros((nbytes,), dtype=numpy.ubyte)

    init_table = [(chr(code),code) for code in range (256)]

    table = {}
    table_get = table.get
    table_clear = table.clear
    table_update = table.update

    sup_code2 = (1<<max_bits) - 2
    next_code = CODESTART
    bits = 9
    max_code = (1<<bits)
    s = ''

    table_update(init_table)

    index = setword(r, 0, bits, CODECLEAR, 1)
    for c in seq:
        s1 = s + c
        if s1 in table:
            s = s1
        else:
            index = setword(r, index, bits, table_get(s), 1)
            table[s1] = next_code
            next_code += 1
            s = c
            if next_code==sup_code2:
                index = setword(r, index, bits, table_get(s), 1)
                index = setword(r, index, bits, CODECLEAR, 1)
                s = ''
                table_clear()
                table_update(init_table)
                next_code = CODESTART
                bits = 9
                max_code = (1<<bits)
            elif next_code==max_code:
                bits += 1
                max_code = (1<<bits)
    if s:
        index = setword(r, index, bits, table_get(s), 1)
    index = setword(r, index, bits, CODEEOI)
    bytes = index//8
    if 8*bytes < index:
        bytes += 1
    return r[:bytes]

def decode_bitarray(bseq):
    """ Decompress Lempel-Ziv-Welch encoded sequence.

    Parameters
    ----------
    bseq : {bitarray, numpy.ndarray}

    Returns
    -------
    seq : str

    See also
    --------
    encode_bitarray
    """
    if isinstance(bseq, numpy.ndarray):
        bseq = bitarray(bseq, endian='little')
    assert bseq.endian ()=='little',repr(bseq.endian())
    read = bseq.toword

    init_invtable = [(code, chr(code)) for code in range (256)]
    table = [chr(code) for code in range(256)] + ['CODECLEAR', 'CODEEOI']
    table_append = table.append
    table_len = table.__len__

    bits = 9
    max_code2 = (1<<bits) - 2
    i = 0
    seq = []
    seq_append = seq.append

    while True:
        code = read(i, bits)
        i += bits
        if code==CODEEOI:
            break
        elif code==CODECLEAR:
            del table[CODESTART:]
            bits = 9
            max_code2 = (1<<bits) - 2
            code = read(i, bits)
            i += bits
            old_str = table[code]
            seq_append(old_str)
            old_code = code
        else:
            l = table_len()
            if code < l:
                s = table[code]
                table_append(old_str + s[0])
                old_str = s
            else:
                old_str = old_str + old_str[0]
                table_append(old_str)

            seq_append(old_str)
            old_code = code

            if l==max_code2:
                bits += 1
                max_code2 = (1<<bits) - 2
    return ''.join(seq)

def decode_bittools(bseq):
    """ Decompress Lempel-Ziv-Welch encoded sequence.

    Parameters
    ----------
    bseq : numpy.ndarray

    Returns
    -------
    seq : str

    See also
    --------
    encode_bittools
    """
    init_invtable = [(code, chr(code)) for code in range (256)]
    table = [chr(code) for code in range(256)] + ['CODECLEAR', 'CODEEOI']
    table_append = table.append
    table_len = table.__len__

    bits = 9
    max_code2 = (1<<bits) - 2
    i = 0
    seq = []
    seq_append = seq.append

    while True:
        code, i = getword(bseq, i, bits)
        if code==CODEEOI:
            break
        elif code==CODECLEAR:
            del table[CODESTART:]
            bits = 9
            max_code2 = (1<<bits) - 2
            code, i = getword(bseq, i, bits)
            old_str = table[code]
            seq_append(old_str)
            old_code = code
        else:
            l = table_len()
            if code < l:
                s = table[code]
                table_append(old_str + s[0])
                old_str = s
            else:
                old_str = old_str + old_str[0]
                table_append(old_str)

            seq_append(old_str)
            old_code = code

            if l==max_code2:
                bits += 1
                max_code2 = (1<<bits) - 2
    return ''.join(seq)


#print 'backend:', default_backend
if default_backend=='bitarray':
    encode = encode_bitarray
    decode = decode_bitarray
    def encode_array(arr):
        return encode_bitarray(arr).toarray ()
if default_backend=='bittools':
    encode = encode_array = encode_bittools
    decode = decode_bittools


def test_lzw():
    for s in ['TOBEORNOTTOBEORTOBEORNOT', '/WED/WE/WEE/WEB/WET'][:0]:
        r = encode (s)
        a = decode (r)
        assert a==s,repr((a,s))

    if 1:
        f = open(__file__)
        s = f.read ()
        f.close ()

        r = encode (s)
        a = decode (r)
        assert a==s
    print('ok')
    import sys
    import os
    import time
    for fn in sys.argv[1:]:
        if not os.path.exists(fn):
            continue
        t0 = time.time()
        f = open(fn, 'rb')
        s = f.read()
        f.close()
        t = time.time()-t0
        print('Reading %s took %.3f seconds, bytes = %s' % (fn, t, len(s)))
        t0 = time.time()
        r = encode(s)
        t = time.time()-t0
        sz = len(r)
        if default_backend=='bitarray':
            sz //= 8
        print('Encoding took %.3f seconds, compress ratio = %.3f, Kbytes per second = %.3f' % (t, len (s)/sz, len(s)/t/1024))
        t0 = time.time()
        s1 = decode(r)
        t = time.time()-t0
        print('Decoding took %.3f seconds, Kbytes per second = %.3f' % (t, (sz/t)/1024))
        assert s1==s

if __name__=='__main__':
    test_lzw()
